/* Creates a compressed image, given a file as an argument.
 * (c)1999 Paul `Rusty' Russell.  GPL.
 *
 * CHANGELOG:
 * * Sat Jun 17 2000 Klaus Knopper <knopper@knopper.net>
 * - Support for reading file from stdin,
 * - Changed Preamble.
 */

#include <stdio.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "compressed_loop.h"
#include "zlib-1.1.3/zlib.h"

#define CLOOP_PREAMBLE "#!/bin/sh\n" "./insmod cloop.o file=$0 && mount -r -t iso9660 /dev/cloop $1\n" "exit $?\n"

struct cb_list
{
	struct cb_list *next;
	size_t size;
	char data[0];
};

void free_cb_list(struct cb_list *cbl)
{
 if(cbl->next) free_cb_list(cbl->next);
 free(cbl);
}

static struct cb_list
	*create_compressed_blocks(int handle, unsigned int blocksize,
		unsigned int *numblocks)
{
	struct cb_list *cbl,**cbp=&cbl;
	unsigned int i,last;

	for(i=0,last=0; !last; i++) {
		unsigned int total=0;
		unsigned long len = blocksize + blocksize/1000 + 12;
		char uncompressed[blocksize], compressed[len];
		memset(uncompressed,0,blocksize);
		while(total<blocksize) {
               		int r=read(handle, uncompressed+total, blocksize-total);
			if(r<=0) { last=1; break; }
			total+=r;
		}
		if (total != blocksize) {
			last=1;
			fprintf(stderr, "Partial read (%i bytes of %u), padding with zeros.\n",
					total, blocksize);
		}
		if (compress(compressed, &len, uncompressed, blocksize)
			    != Z_OK) {
			fprintf(stderr, "Error compressing block %i\n", i);
			if(cbl) free_cb_list(cbl);
			return NULL;
		}

		if ((*cbp = malloc(sizeof(struct cb_list)+len))==NULL) {
			fprintf(stderr, "Out of memory allocating block ptrs.\n");
			if(cbl) free_cb_list(cbl);
			return NULL;
		}

		fprintf(stderr, "Block %u size %lu\n", i, len);
		(*cbp)->size = len;
		memcpy((*cbp)->data, compressed, len);
		(*cbp)->next=NULL;
		cbp=&((*cbp)->next);
		memset(uncompressed,0,blocksize);
	} /* for */
	*numblocks=i;
	return cbl;
}

int main(int argc, const char *argv[])
{
	int input;
	unsigned int blocksize;
	struct cloop_head head;
	unsigned int numblocks;
	unsigned int bytes_so_far;
	unsigned int i;
	struct cb_list *compressed_blocks,*cbp;

	if (argc != 3) {
		fprintf(stderr, "Usage: %s filename blocksize(bytes).\n",argv[0]);
		fprintf(stderr, "Use '-' as filename for stdin.\n");
		return 1;
	}

	blocksize = atoi(argv[2]);
	if (blocksize == 0 || blocksize % 512 != 0) {
		fprintf(stderr, "Blocksize must be a multiple of 512.\n");
		return 1;
	}

	if (sizeof(CLOOP_PREAMBLE) > CLOOP_HEADROOM) {
		fprintf(stderr, "Preamble (%u chars) > headroom (%u)\n",
			sizeof(CLOOP_PREAMBLE), CLOOP_HEADROOM);
		return 1;
	}
		
	input=strcmp(argv[1],"-")==0?dup(fileno(stdin)):open(argv[1], O_RDONLY);

	if (input < 0) {
		perror("Opening input");
		return 1;
	}


	compressed_blocks =
		 create_compressed_blocks(input, blocksize, &numblocks);

	close(input);

	memcpy(head.preamble, CLOOP_PREAMBLE, sizeof(CLOOP_PREAMBLE));
	head.block_size = htonl(blocksize);
	head.num_blocks = htonl(numblocks);

	fprintf(stderr,
		"Block size %u, number of blocks %u.\n", blocksize, numblocks);

	bytes_so_far = sizeof(head) + sizeof(u_int32_t) * (numblocks + 1);

	/* Write out head... */
	write(STDOUT_FILENO, &head, sizeof(head));

	if (!compressed_blocks)
		return 1;

	/* Write offsets */
	for (i=0,cbp=compressed_blocks; i < numblocks+1; i++) {
		u_int32_t tmp;
		tmp = htonl(bytes_so_far);
		write(STDOUT_FILENO, &tmp, sizeof(tmp));
		if(cbp) { bytes_so_far += cbp->size; cbp=cbp->next; }
	}

	/* Now write blocks. */
	for (i = 0, cbp=compressed_blocks; i < numblocks; i++) {
		if (write(STDOUT_FILENO, cbp->data, cbp->size)
		    != cbp->size) {
			perror("writing block");
			free_cb_list(compressed_blocks);
			return 1;
		}
		cbp=cbp->next;
	}
	free_cb_list(compressed_blocks);
	return 0;
}
